/*
 * CLEP-CIDRE :-))
 * ---------------
 *  Cercle Lumineux Electronique Programmé sous bouteilles de CIDRE
 *  ou
 *  Circuit Lumineux Elliptique Programmé sous bouteilles de CIDRE
 *
 *  Ce programme allume les leds selon deux modes:
 *    . mode "normal" (ou "horloge"): 
 *      ..  les heures (0 à 11 et 12 à 23) sont affichées
 *          au moyen des 12 leds "heures"
 *      ..  les minutes (0, 5, 10, .. 55) sont affichées
 *          au moyen des 12 leds "minutes"
 *      ..  Les secondes utilisent les leds des minutes en les faisant clignoter
 *          pendant 5 secondes chacune (celle de 0 minutes pour les secondes 0 à 4,
 *          celle de 5 minutes pour les secondes de 5 à 9 etc.). La led des minutes actuelles ne clignote pas
 *    . mode "random" (ou "aléatoire"):
 *      ..  les 24 leds (heures et minutes) clignotent de façon tout-a-fait aléatoire
 *  
 *  Le programme commute entre ces deux modes selon le schéma suivant en fonction des minutes actuelles
 *    . Minutes multiples de 5 (M5 = 0,5,10,..): 30sec en mode normal et 0 seconde en mode aléatoire
 *    . Ecart entre minutes actuelles et dernier M5 = 1 (1,6,11,..): 24sec en mode normal et 6 secondes en mode aléatoire
 *    . Ecart entre minutes actuelles et dernier M5 = 2 (2,7,12,..): 18sec en mode normal et 12 secondes en mode aléatoire
 *    . Ecart entre minutes actuelles et dernier M5 = 3 (3,8,13,..): 12sec en mode normal et 18 secondes en mode aléatoire
 *    . Ecart entre minutes actuelles et dernier M5 = 4 (4,9,14,..): 6sec en mode normal et 24 secondes en mode aléatoire
 *    
 *  L'allumage des leds n'a lieu que lorsqu'une présence est détectée et ce pendant une durée fixe (5 minutes) pour autant
 *  qu'une nouvelle présence ne soit détectée, auquel cas la période est remise à zéro.
 * 
 *  Copyright 2019: R. Van de Moosdijk
 * 
 */

/*
 *********************
 * Include libraries
 *********************
 */

//  Wire library for I2C
#include <Wire.h>
//  RTC library for using the RTC module
#include <RTClib.h>
//  74HC595 shiftregisters
#include  <ShiftRegister74HC595.h>

/*
 * **********  DECLARATIONS  *************
 */


/*
 ********************
 *  Global variables
 ********************
 */

#define HLED_COUNT  12            //  Number of "hour" LEDs
#define MLED_COUNT  12            //  Nombre de "minute" LEDs
#define MODE_COUNT  3             //  Number of modes
#define NORMAL_MODE_OFFSET  0     //  Offset for the mode led "normal"
#define RAND_MODE_OFFSET    1     //  Offset for the mode led "random"
#define VISIT_MODE_OFFSET   2     //  Offset for the mode led "visitor"
#define RAND_PATT_H_MIN     0     //  Valeur min pour les bits "hours"
#define RAND_PATT_M_MIN     0     //  Valeur min pour les bits "minutes"
#define RAND_PATT_H_MAX     4096  //  Valeur max pour les bits "hours"
#define RAND_PATT_M_MAX     4096  //  Valeur max pour les bits "minutes"
#define NR_OF_SR_H        2     // Nr of 75HC595 shift registers for hour leds
#define NR_OF_SR_M        2     // Nr of 75HC595 shift registers for minute leds

#define DS_M              4     //  Pin nr for Data line for first hour shift register
#define ST_CP_M           5     //  Pin nr for ST_CP (latch) for the hour shift registers
#define SH_CP_M           6     //  Pin nr for SH_CP (reg clock) for the hour shift registers


#define DS_H              7     //  Pin nr for Data line for first minute shift register
#define ST_CP_H           8     //  Pin nr for ST_CP (latch) for the minute shift registers
#define SH_CP_H           9     //  Pin nr for SH_CP (reg clock) for the minute shift registers


//#define         TEST_RUN              //  Uncomment this line if in test mode
//#define         NO_RTC                //  Uncomment this line if no RTC connected
//#define         VISITOR_MODE          //  Mode switching depends on visitMode
                                        //  i.e. visitor detection does not influence
                                        //  the leds' on/off switching
//
//  >>> Mode variables
//
boolean normalMode      = true;         //  Controls the normal led switching timing
char    fromMode        = 'N';          //  Indicates from wich mode the swicthing is requested
boolean visitMode       = true;         //  Applies visitor params to timing
boolean randMode        = false;        //  Switch all leds randomly
boolean switchModeNeeded= false;        //  When set, mode switches between normal and random
boolean firstReadClock  = false;

//
// >>> Timing variables
//

unsigned  long  curMillis = 0;

//  Read clock timing
unsigned  long  prevReadClockMillis = 0;
unsigned  long  readClockInterval   = 1000;               //  5 seconds

//  Blinking
unsigned  long  secBlinkInterval    = 500;                //  500 msec
unsigned  long  prevSecBlinkMillis  = 0;

//  NORMAL mode
unsigned  long  normalModePrevMillis= 0;
unsigned  long  normalModeDefInt    = 30000;              //  Default = 30 seconds
unsigned  long  normalModeInterval  = normalModeDefInt;   //  Current normal mode interval
long            normalModeIntMin    = 20000;              //  20 seconds minimum
long            normalModeIntMax    = 60000;              //  1 minute max
long            normalModeIntMinVis = 5000;               //  Minimum when visitor mode active
long            normalModeIntMaxVis = 10000;              //  Maximum when visitor mode active
long            normalModeStep      = 6000;               //  normalModeInterval 
                                                          //  = normalModeInterval - 
                                                          //  (#min from last multiple of 5min)*normalModeStep
                                                          //  randModeInterval 
                                                          //  = (#min from last multiple of 5min)*normalModeStep

//  RANDOM mode
unsigned  long  randModePrevMillis  = 0;
unsigned  long  randModeInterval    = 30000;              //  Default = 10 seconds
unsigned  long  randBlinkPrevMillis = 0;                  //  Last time when switching leds randomly
unsigned  long  randBlinkInterval   = 1000;               //  Random switching interval
long            randModeIntMin      = 20000;              //  30 seconds minimum
long            randModeIntMax      = 60000;              //  1 minute max
long            randModeIntMinVis   = 3000;               //  Minimum when visitor mode active
long            randModeIntMaxVis   = 8000;               //  Maximum when visitor mode active

//  VISITOR mode
unsigned  long  visitModePrevMillis = 0;
unsigned  long  visitModeInterval   = 300000;             //  Default = 10 seconds
long            visitModeIntMin     = 30000;              //  30 seconds min
long            visitModeIntMax     = 60000;              //  1 minute max;

/*
 ****************************
 *  Données "LEDs"
 ****************************
 */

const int ledBI =  LED_BUILTIN;
const int modeLeds[MODE_COUNT] = {
        10,    //  Normal mode
        11,    //  Random mode
        12     //  Visitor mode
};

//
//  Détecteur de présence
const int visitorDetPin   = 2;                    //  Presence detection connected to pin 2

#ifndef NR_OF_SR_H
//
//  Table des numéros de sorties LED "heure"
const int   hourLedsPos[HLED_COUNT] = {  
  //  Correspondance "heure" (0, 1, ..) vers "numéro de sortie"
                                    22,    //  h=0 ou 12
                                    24,    //  h=1 ou 13
                                    26,    //  h=2 ou 14
                                    28,    //  h=3 ou 15
                                    30,    //  h=4 ou 16
                                    32,    //  h=5 ou 17
                                    34,    //  h=6 ou 18
                                    36,    //  h=7 ou 19
                                    38,    //  h=8 ou 20
                                    40,    //  h=9 ou 21
                                    42,    //  h=10 ou 22
                                    44     //  h=11 ou 23
                                  };

#else
byte  clockValues[24] = { //  Byte values for hour and minute leds
                          //  12 out of the 16 bits are used
  0,1,    //  0000,00000001
  0,2,    //  0000,00000010
  0,4,    //  0000,00000100
  0,8,    //  0000,00001000
  0,16,   //  0000,00010000
  0,32,   //  0000,00100000
  0,64,   //  0000,01000000
  0,128,  //  0000,10000000
  1,0,    //  0001,00000000
  2,0,    //  0010,00000000
  4,0,    //  0100,00000000
  8,0     //  1000,00000000
};

//  Output arrays for hour and minute leds
uint8_t regOutputHour [2] = {0,0};
uint8_t regOutputMin  [2] = {0,0};

#endif


#ifndef NR_OF_SR_M
//
//  Table des numéros de sorties LED "minute"
const int   minuteLedsPos[MLED_COUNT] = {  
  //  Correspondance "minute" (0, 5, 10, ..) vers "numéro de sortie"
                                    23, //  m=0   ..  4
                                    25, //  m=5   ..  9
                                    27, //  m=10  ..  14
                                    29, //  m=15  ..  19
                                    31, //  m=20  ..  24
                                    33, //  m=25  ..  29
                                    35, //  m=30  ..  34
                                    37, //  m=35  ..  39
                                    39, //  m=40  ..  44
                                    41, //  m=45  ..  49
                                    43, //  m=50  ..  54
                                    45  //  m=55  ..  59
                                  };
#endif

#ifndef NR_OF_SR_H
int curHourLed    = hourLedsPos[0];         //  Moves to next one every 1 hour
                                            //  and loop after 12 hours
#endif

int curHourOffset   = 0;
int curMinuteOffset = 0;
int curSecondOffset = 0;

int curMinuteLed  = 0;       //  Moves to next one every 5 minutes
                                            //  and loop after 60 minutes
int curSecondLed  = 0;       //  Moves to next one every 5 seconds
                                            //  and loop after 60 seconds
                                            //  Blinks during 5 seconds
                                            //  unless curSecondLed == curMinuteLed
int lastSecondLed = curSecondLed;

boolean curBlinkState = true;               //  The current "on/off" state of the second blinking led

/*
 ****************************
 *  Shift registers
 ****************************
 */
#ifdef  NR_OF_SR_H
ShiftRegister74HC595 srH (NR_OF_SR_H, DS_H, SH_CP_H, ST_CP_H); 
#endif
#ifdef  NR_OF_SR_M
ShiftRegister74HC595 srM (NR_OF_SR_M, DS_M, SH_CP_M, ST_CP_M); 
#endif
//

/*
 ****************************
 *  Données "RTC"
 ****************************
 */

RTC_DS1307 rtc;     //  Create RTC object
int rtc_hour  = 0;
int rtc_min   = 1;
int rtc_sec   = -5;

/*
 * **********  END OF DECLARATIONS  *************
 */


/*
 *****************
 *     SET-UP    *
 *****************
 */
 
void setup() {
  
  Serial.begin(57600);
  Serial.flush();

  srH.setAllLow();

  pinMode(ledBI,OUTPUT);
  pinMode(visitorDetPin, INPUT);

#ifndef NR_OF_SR_H
  //  Définition sorties LED "heures"
  for (int h = 0; h < HLED_COUNT; h++){
    pinMode(hourLedsPos[h], OUTPUT);
  }
#else
  pinMode(DS_H, OUTPUT);
  pinMode(ST_CP_H, OUTPUT);
  pinMode(SH_CP_H, OUTPUT);
#endif
#ifndef NR_OF_SR_M
  //  Définition sorties LED "minutes" (and "blinking seconds")
  for (int m = 0; m < MLED_COUNT; m++){
    pinMode(minuteLedsPos[m], OUTPUT);
  }
#else
  pinMode(DS_M, OUTPUT);
  pinMode(ST_CP_M, OUTPUT);
  pinMode(SH_CP_M, OUTPUT);
#endif

  //  Définition sorties LED "modes"
  for (int m = 0; m < MODE_COUNT; m++){
    pinMode(modeLeds[m], OUTPUT);
  }

//
//  Search RTC module
  if (! rtc.begin()) {
    Serial.println("Couldn't find RTC");
    while (1);
  }

//
//  Adjust RTC time from compile time
  if (! rtc.isrunning()) {
    Serial.println("RTC is NOT running!");
    // following line sets the RTC to the date & time this sketch was compiled
     rtc.adjust(DateTime(F(__DATE__), F(__TIME__)));
    // This line sets the RTC with an explicit date & time, for example to set
    // January 21, 2014 at 3am you would call:
    // rtc.adjust(DateTime(2014, 1, 21, 3, 0, 0));
  }

  digitalWrite(ledBI,LOW);

//  Initialize seed for random function
  randomSeed(analogRead(0));

  while( rtc_sec != 0){
    readClock();
    delay(100);
  }
  
  setClockLeds();
  
  curMillis = millis();
  calculateModeIntervals();
  normalModePrevMillis  = curMillis;
  randModePrevMillis    = curMillis;
  Serial.println("Démarrage ");
  printMode('S');

//  Attach interrupt routine to deector pin
  attachInterrupt(0, visitorDetected, RISING);

} //  SET-UP


/*
 *****************
 *     LOOP      *
 *****************
*/
 
void loop() {

  curMillis = millis();
  
//  
//  READ CLOCK

  if (curMillis < prevReadClockMillis){prevReadClockMillis = curMillis;}
  if (curMillis >= prevReadClockMillis + readClockInterval){
    //  Read RTC values
    prevReadClockMillis = curMillis;
    readClock();
//
//  Switch hour and minute leds according to rtc values
#ifndef NR_OF_SR_H
    curHourLed    = hourLedsPos[hourToOffset(rtc_hour)];
#endif
#ifndef NR_OF_SR_M
    curMinuteLed  = minuteLedsPos[minToOffset(rtc_min)];
    curSecondLed  = minuteLedsPos[secToOffset(rtc_sec)];
#endif

//  Switch leds on
    setClockLeds();

  }// End of read clock

//  
//  MANAGE MODES

  if(switchModeNeeded){
    switchMode(fromMode);
  }
  setModeLeds();

//  Visitor mode
  if(visitMode){
    if  (curMillis < visitModePrevMillis){visitModePrevMillis = curMillis;}
    if  (curMillis >= visitModePrevMillis + visitModeInterval){
//      visitMode = false;
      fromMode  = 'E';
#ifdef VISITOR_MODE
      switchModeNeeded = true;
#endif
    }
  }


//
//  NORMAL mode
  if(normalMode){
    if  (curMillis < normalModePrevMillis){normalModePrevMillis = curMillis;}
    if  (curMillis > normalModePrevMillis + normalModeInterval){   
      fromMode = 'N';
      switchModeNeeded = true;
      readClock();
      calculateModeIntervals();
    }
    else{
      if  (curMillis < prevSecBlinkMillis){prevSecBlinkMillis = curMillis;}
      if  (curMillis >= prevSecBlinkMillis + secBlinkInterval){
        prevSecBlinkMillis = curMillis;
        setSecondLed();
      }
    }
  }


//
//  RANDOM mode
  if(randMode){
    if  (curMillis < randModePrevMillis){randModePrevMillis = curMillis;}
    if  (curMillis > randModePrevMillis + randModeInterval){
      fromMode = 'R';
      switchModeNeeded = true;
      readClock();
      calculateModeIntervals();
    }
    else{
      if  (curMillis < randBlinkPrevMillis){randBlinkPrevMillis = curMillis;}
      if  (curMillis >= randBlinkPrevMillis + randBlinkInterval){
        randBlinkPrevMillis = curMillis;
        setRandomLeds();
      }
    }
  }

} //  End of LOOP


/*
 *****************
 *   FUNCTIONS   *
 *****************
 */

//
//  Read rtc values
void  readClock(){
#ifdef  NO_RTC
    rtc_hour = 14;
    rtc_min = 37;
    rtc_sec += readClockInterval/1000;
    if(rtc_sec >= 60){rtc_sec = 0;}
#else
    DateTime  now = rtc.now();
    rtc_hour  = now.hour();
    rtc_min   = now.minute();
    rtc_sec   = now.second();
#endif
}


//
//  Hours to offset conversion
int hourToOffset(int hint){// hint between 0 and 23
  return hint-((hint/12)*12);
}

//
//  Minutes to offset conversion
int minToOffset(int mint){//  mint between 0 and 59
  return mint/5;
}

//
//  Seconds to offset conversion
int secToOffset(int sint){//  sint between 0 and 59
  return sint/5;
}


//
//  Split one integer value into an array of two bytes
//  (LSB at index 1, MSB at index 0)
void  splitIntToBytes(int intVal, uint8_t outByteArray[2]){
  outByteArray[0] = intVal & 0xFF;
  outByteArray[1] = (intVal >> 8) & 0xFF;
}

//
//  Set the hour and minute leds according to current led pin nrs
void setClockLeds(){
  if(!randMode){
    //
    //  Switch hour led
#ifndef NR_OF_SR_H
      for (int ho = 0; ho < HLED_COUNT; ho++){
        digitalWrite(hourLedsPos[ho], LOW);
      }
#endif
#ifndef  VISITOR_MODE
      if(visitMode){
#endif
#ifndef NR_OF_SR_H
        digitalWrite(curHourLed, HIGH);
#else
  curHourOffset = hourToOffset(rtc_hour); 
  regOutputHour[0] =  clockValues[curHourOffset*2 + 1];
  regOutputHour[1] =  clockValues[curHourOffset*2];
//  Output the minute leds to the shift registers
  srH.setAll(regOutputHour);
#endif

#ifndef  VISITOR_MODE
        }
#endif
    //
    //  Switch minute led
#ifndef NR_OF_SR_M    
      for (int mi = 0; mi < MLED_COUNT; mi++){
          digitalWrite(minuteLedsPos[mi], LOW);
      }
#endif

#ifndef  VISITOR_MODE
      if(visitMode){
#endif

#ifndef NR_OF_SR_M
        digitalWrite(curMinuteLed, HIGH);
#else
  curMinuteOffset = minToOffset(rtc_min); 
  regOutputMin[0] =  clockValues[curMinuteOffset*2 + 1];
  regOutputMin[1] =  clockValues[curMinuteOffset*2];
//  Output the minute leds to the shift registers
  srM.setAll(regOutputMin);
#endif

#ifndef  VISITOR_MODE    
        }
#endif
  }
}

//
//  Switch (on/off) the minute leds as "second led blinking"
void  setSecondLed(){
#ifndef  VISITOR_MODE
  if(visitMode){
#endif
#ifndef NR_OF_SR_M
    if(curMinuteLed != curSecondLed){
      digitalWrite(curSecondLed, curBlinkState);
    }
    else{
      if(curMinuteLed != lastSecondLed){
        digitalWrite(lastSecondLed, LOW);   
      }
    }
#else
    curSecondOffset = secToOffset(rtc_sec);
    curMinuteOffset = minToOffset(rtc_min);
    if(curSecondOffset != curMinuteOffset){
      srM.set(curSecondOffset, curBlinkState);
    }
#endif
#ifndef  VISITOR_MODE
  }
#endif
#ifndef NR_OF_SR_H
  lastSecondLed = curSecondLed;
#endif
  curBlinkState = !curBlinkState;
}

//
//  Set leds given by random patterns
void  setRandomLeds(){
  int randomPatternHours    = 0;
  int randomPatternMinutes  = 0;
#ifndef  VISITOR_MODE
  if(visitMode){
#endif
    randomPatternHours    = random(RAND_PATT_H_MIN,RAND_PATT_H_MAX);
    randomPatternMinutes  = random(RAND_PATT_M_MIN,RAND_PATT_M_MAX);
#ifndef  VISITOR_MODE
  }
#endif
#ifndef NR_OF_SR_H
  int msk = 1;
  boolean ledState = false;
  for (int h = 0; h < HLED_COUNT; h++){
      ledState = randomPatternMinutes & msk;
      digitalWrite(hourLedsPos[h], ledState);
      msk = msk << 1;
  }
#else
  splitIntToBytes(randomPatternHours, regOutputHour);
  srH.setAll(regOutputHour);
#endif
#ifndef NR_OF_SR_M
  msk = 1;
  ledState = false;
  for (int m = 0; m < MLED_COUNT; m++){
      ledState = randomPatternMinutes & msk;
      digitalWrite(minuteLedsPos[m], ledState);
      msk = msk << 1;
  }
#else
  splitIntToBytes(randomPatternMinutes, regOutputMin);
  srM.setAll(regOutputMin);
#endif
}

//
//  Set the mode leds according to current modes
void  setModeLeds(){
#ifndef NR_OF_SR_H
  digitalWrite(modeLeds[NORMAL_MODE_OFFSET],normalMode);
  digitalWrite(modeLeds[RAND_MODE_OFFSET],randMode);
  digitalWrite(modeLeds[VISIT_MODE_OFFSET],visitMode);
#endif
}

/*
 *  >>>  Detector ISR
*/
void  visitorDetected(){
  curMillis = millis();
  byte state = digitalRead(visitorDetPin);
  if (state != 0){  //  Signal haut   
    visitModePrevMillis = curMillis;
    if(!visitMode){
#ifdef  VISITOR_MODE
      visitModeInterval = random(visitModeIntMin,visitModeIntMax);
#endif
      visitMode   = true;
    }
    fromMode  = 'D';
#ifdef  VISITOR_MODE
    switchModeNeeded = true;
#endif
  }
}

//
//  Switch between NORMAL and RANDOM modes
void  switchMode(char fm){
    long  minInterval = 10000;
    long  maxInterval = 30000;
    normalMode  = !normalMode;
    randMode    = !randMode;
    if(normalMode){
#ifdef VISITOR_MODE
      if (visitMode){
        minInterval = normalModeIntMinVis;
        maxInterval = normalModeIntMaxVis;
      }
      else{
        minInterval = normalModeIntMin;
        maxInterval = normalModeIntMax;
      }
      normalModeInterval  = random(minInterval,maxInterval);
#endif
      normalModePrevMillis = curMillis;
    }

    if(randMode){
#ifdef  VISITOR_MODE
      if (visitMode){
        minInterval = randModeIntMinVis;
        maxInterval = randModeIntMaxVis;
      }
      else{
        minInterval = randModeIntMin;
        maxInterval = randModeIntMax;
      }
      randModeInterval  = random(minInterval,maxInterval);
#endif
      randModePrevMillis = curMillis;
      randBlinkPrevMillis = curMillis;
    }
    switchModeNeeded = false;

    printMode(fm);
}

//
//  Calculate mode intervals depending on the current minutes
void  calculateModeIntervals(){
#ifndef  VISITOR_MODE
  randModeInterval = (rtc_min%5)*normalModeStep;
  normalModeInterval = normalModeDefInt - randModeInterval;
#else
  normalModeInterval  = random(normalModeIntMin,normalModeIntMax);
  randModeInterval    = random(randModeIntMin,randModeIntMax);
#endif
}


//
//  Printing routines
//
//  Print the current tme (H, M, S) as returned by the RTC
void  printTime(){
//  Serial.print(curMillis);
//  Serial.print(" msec ");
  Serial.print(rtc_hour, DEC);
  Serial.print(':');
  Serial.print(rtc_min, DEC);
  Serial.print(':');
  Serial.print(rtc_sec, DEC);
  Serial.print(" ");
}

//
//  Print the current led pin numbers
void  printLed(){
#ifndef NR_OF_SR_M
    Serial.print(curHourLed);
    Serial.print(" ");
    Serial.print(curMinuteLed);
    Serial.print(" ");
    Serial.print(curSecondLed);
    Serial.print(" ");
    Serial.println(lastSecondLed);
#endif
}

//
//  Print the mode values
void  printMode(char x){
    printTime();
    Serial.print("Switched from ");
    Serial.print(x);
    Serial.print(" ");
    if (visitMode){
      Serial.print("Visitor for ");
      Serial.print(visitModeInterval, DEC);
      Serial.print(" msec  ");
    }
    if (normalMode){
      Serial.print("Normal for ");
      Serial.print(normalModeInterval, DEC);
      Serial.print(" msec");
    }
    if (randMode){
      Serial.print("Random for ");
      Serial.print(randModeInterval, DEC);
      Serial.print(" msec");
      Serial.print(" Switching interval :");
      Serial.print(randBlinkInterval, DEC);
      Serial.print(" msec");
    }
    Serial.println();
}
